/* ---------------------------------------------------------------------------

  This file is part of the ``rvsynth'' package (client side).
  Copyright (R) 2018  Fondazione Bruno Kessler

  Some code pieces were from Kaspersky Security System Demos

  Author: Chun Tian
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <orbit/orbit.h>

/*
 * This header file was generated from the idl
 */

#include "monitor.h"
#include "toolkit.h" /* ie. etk_abort_if_exception() */

/**
 * Print current time to standard output.
 */
static void print_time()
{
    char buf[20];
    size_t buf_len;
    struct timeval now;
    int milli;

    gettimeofday(&now, NULL);
    milli = now.tv_usec / 1000;
    buf_len = strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", localtime(&now.tv_sec));
    if (buf_len > 0)
        printf("%s.%03d ", buf, milli);
}

static CORBA_ORB global_orb = CORBA_OBJECT_NIL; /* global orb */
 
/* Is called in case of process signals. it invokes CORBA_ORB_shutdown()
 * function, which will terminate the processes main loop.
 */
static void client_shutdown (int sig)
{
  CORBA_Environment  local_ev[1];
  CORBA_exception_init(local_ev);

  if (global_orb != CORBA_OBJECT_NIL) {
    CORBA_ORB_shutdown (global_orb, FALSE, local_ev);
    etk_abort_if_exception (local_ev, "caught exception");
  }
}

/* Inits ORB @orb using @argv arguments for configuration. For each
 * ORBit options consumed from vector @argv the counter of @argc_ptr
 * will be decremented. Signal handler is set to call
 * echo_client_shutdown function in case of SIGINT and SIGTERM
 * signals.  If error occures @ev points to exception object on
 * return.
 */
static void
client_init (int               *argc_ptr,
	     char              *argv[],
             CORBA_ORB         *orb,
             CORBA_Environment *ev,
	     Monitor_Service   *service)
{
  CosNaming_NamingContext name_service = CORBA_OBJECT_NIL;
  gchar *id[] = {"FBK", "Monitor", "Service", NULL};

  CORBA_exception_init(ev);
  /* init signal handling */
  signal(SIGINT,  client_shutdown);
  signal(SIGTERM, client_shutdown);

  /* create Object Request Broker (ORB) */
  (*orb) = CORBA_ORB_init(argc_ptr, argv, "orbit-local-orb", ev);
  if (etk_raised_exception(ev)) return;

  g_print ("Resolving service reference from name-service with id \"%s/%s/%s\"\n",
	   id[0], id[1], id[2]);

  name_service = etk_get_name_service (global_orb, ev);
  etk_abort_if_exception(ev, "failed resolving name-service");

  *service = (Monitor_Service) etk_name_service_resolve (name_service, id, ev);
  etk_abort_if_exception(ev, "failed resolving service at name-service");

}

/* Releases @servant object and finally destroys @orb. If error
 * occures @ev points to exception object on return.
 */
static void client_cleanup (CORBA_ORB          orb,
			    CORBA_Object       service,
			    CORBA_Environment *ev)
{
  /* releasing managed object */
  CORBA_Object_release(service, ev);
  if (etk_raised_exception(ev)) return;

  /* tear down the ORB */
  if (orb != CORBA_OBJECT_NIL) {
    /* going to destroy orb.. */
    CORBA_ORB_destroy(orb, ev);
    if (etk_raised_exception(ev)) return;
  }
}

static int heartbeat (Monitor_Service service,
		      CORBA_Environment *ev,
		      int p_or_q, int index)
{
  Monitor_Output output;
  int retval;
  char *constraint;

  if (p_or_q)
    constraint = "p";
  else
    constraint = "q";

  /* using the heartbeat method in the Monitor_Service object 
   * this is defined in the echo.h header, compiled from
   * monitor.idl */
  output = Monitor_Service_control(service, constraint, index, ev);

  print_time();
  g_print("[Command] HEARTBEAT(%s) ... ", (p_or_q ? "p" : "q"));

  switch (output) {
  case Monitor_Yes:
    g_print("SUCCESS\n");
    retval = 1; /* reset then */
    break;
  case Monitor_No:
    g_print("FAIL   \n");
    retval = 1; /* reset then */
    break;
  case Monitor_Unknown:
    g_print("OK     \n");
    retval = 0; /* continue the loop */
    break;
  default:
    g_print("ERROR\n");
    exit(-1);
  }
  if (etk_raised_exception (ev)) return -1;

  return retval;
}

static int reset (Monitor_Service service,
		  CORBA_Environment *ev, int index)
{
  print_time();
  printf("[Command] RESET\n");

  /* reset (stop) the monitor (one way function) */
  Monitor_Service_restart(service, index, ev);
  if (etk_raised_exception (ev)) return -1;
  
  return 0;
}

Monitor_Service   service = CORBA_OBJECT_NIL;
CORBA_Environment ev[1];

void test (int index) {
  int i, p_or_q;
  for (i = 0; i < 16; i++) {
    p_or_q = rand() % 2;
    if (heartbeat(service, ev, p_or_q, index) != 0)
      (void) reset(service, ev, index);
    usleep(200000);
  }
}

void test_shared(int index, char* constraint, int retval[]) {
  Monitor_Output output;

  output = Monitor_Service_control (service, constraint, index, ev);
  switch (output) {
  case Monitor_Yes:
    g_print("SUCCESS\t");
    retval[index] = 1; /* reset then */
    break;
  case Monitor_No:
    g_print("FAIL   \t");
    retval[index] = 1; /* reset then */
    break;
  case Monitor_Unknown:
    g_print("OK     \t");
    retval[index] = 0; /* continue the loop */
    break;
  default:
    g_print("ERROR\n");
    exit(-1);
  }
}

void test_all () {
  int i, p_or_q;
  char *constraint;
  
  static int retval[3]; /* holding reset signals */

  for (i = 0; i < 16; i++) {
    p_or_q = rand() % 2;

    if (p_or_q)
      constraint = "p";
    else
      constraint = "q";

    print_time();
    g_print("[Command] HEARTBEAT(%s) ... ", (p_or_q ? "p" : "q"));

    test_shared(0, constraint, retval);
    test_shared(1, constraint, retval);
    test_shared(2, constraint, retval);
    printf("\n");
 
    if (retval[0] || retval[1] || retval[2]) {
      print_time();
      printf("[Command] RESET "); /* one way function */
      if (retval[0]) {
	Monitor_Service_restart (service, 0, ev);
	printf("0 ");
      }
      if (retval[1]) {
	Monitor_Service_restart (service, 1, ev);
	printf("1 ");
      }
      if (retval[2]) {
	Monitor_Service_restart (service, 2, ev);
	printf("2 ");
      }
      printf("\n");
    }
    usleep(200000);
  }
}

/*
 * main 
 */
int main (int argc, char* argv[])
{
  /* Startup messages */
  g_print("*** This is a sample monitor client of NuRV runtime monitor based on nuXmv\n");
  g_print("*** Copyright (C) 2018, Fondazione Bruno Kessler\n\n");
  g_print("*** Usage: %s -ORBInitRef NameService=IOR:... \n\n", argv[0]);

  client_init (&argc, argv, &global_orb, ev, &service);
  etk_abort_if_exception(ev, "init failed");

  srand(time(NULL));

  printf("\nTest Part 1: (p U q)\n");
  test(0);
  
  printf("\nTest Part 2: G (p -> X q)\n");
  test(1);

  printf("\nTest Part 3: F (Y p & p & X q)\n");
  test(2);

  printf("\nTest Part 4: all together -- (p U q),\tG (p -> X q),\tF (Y p & p & X q)\n");
  test_all();
  
  client_cleanup (global_orb, service, ev);
  etk_abort_if_exception(ev, "cleanup failed");

  exit (0);
}
