#define FMI_VERSION 2

#include <fmi2Functions.h>
#include <stdlib.h>
#include <string.h>
#include "monitor.c"

static const fmi2String monitor_uuid = "{2427283A-F8B0-11ED-BF85-ACDE48001122}";

struct monitor_fmu {
  char *instanceName;    /* const */  int current_loc;       /* in */
  monitor_input_t input; /* in */
  RV_value output;       /* out */
  int soft_reset;        /* in */
  fmi2CallbackFunctions *functions;
};

/* FMI functions */
const char* fmi2GetTypesPlatform () { return fmi2TypesPlatform; }

const char* fmi2GetVersion () { return fmi2Version; }

fmi2Status fmi2SetDebugLogging (fmi2Component _instance, fmi2Boolean loggingOn,
                                size_t nCategories, const fmi2String categories[])
  { return fmi2OK; }

fmi2Component fmi2Instantiate (fmi2String instanceName,
                               fmi2Type fmuType,
                               fmi2String fmuGUID,
                               fmi2String fmuResourceLocation,
                               const fmi2CallbackFunctions *functions,
                               fmi2Boolean visible,
                               fmi2Boolean loggingOn)
{
  struct monitor_fmu *_mon;

  /* 1. check GUID */
  if (strcmp(fmuGUID, monitor_uuid) != 0) { return NULL; }
  /* 2. check FMU type (only CoSimulation is supported) */
  if (fmuType != fmi2CoSimulation) { return NULL; }
  if (functions && functions->allocateMemory) {
    _mon = (struct monitor_fmu *)
             functions->allocateMemory(1, sizeof(struct monitor_fmu));
    if (_mon == NULL) return NULL;
    memset((void *)_mon, 0, sizeof(struct monitor_fmu));
    _mon->functions = (fmi2CallbackFunctions *)functions;
  } else {
    _mon = (struct monitor_fmu *)calloc(1, sizeof(struct monitor_fmu));
  }

  if (functions && functions->allocateMemory) {
    _mon->instanceName = (char *)
      functions->allocateMemory(strlen(instanceName) + 1, sizeof(char));
    strcpy(_mon->instanceName, instanceName);
  } else {
    _mon->instanceName = strdup(instanceName);
  }
  _mon->current_loc = 1; /* initial location */
  return (fmi2Component)_mon;
}

fmi2Status fmi2SetupExperiment (fmi2Component _instance,
                                fmi2Boolean toleranceDefined,
                                fmi2Real tolerance, fmi2Real startTime,
                                fmi2Boolean stopTimeDefined,
                                fmi2Real stopTime)
{ return fmi2OK; }

fmi2Status fmi2EnterInitializationMode (fmi2Component _instance)
{ return fmi2OK; }

fmi2Status fmi2ExitInitializationMode (fmi2Component _instance)
{ return fmi2OK; }

fmi2Status fmi2Terminate (fmi2Component _instance) { return fmi2OK; }

fmi2Status fmi2Reset (fmi2Component _instance)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  memset((char *)_mon, 0, sizeof(struct monitor_fmu));
  _mon->current_loc = 1;
  return fmi2OK;
}

void fmi2FreeInstance (fmi2Component _instance)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  const fmi2CallbackFunctions *functions = _mon->functions;
  if (functions && functions->freeMemory) {
    functions->freeMemory(_mon->instanceName);
    functions->freeMemory(_mon);
  } else {
    free(_mon->instanceName);
    free(_mon);
  }
}

fmi2Status fmi2GetReal (fmi2Component _instance, const fmi2ValueReference vr[],
                        size_t nvr, fmi2Real value[])
{ return fmi2Warning; }

fmi2Status fmi2GetInteger (fmi2Component _instance, const fmi2ValueReference vr[],
                           size_t nvr, fmi2Integer value[])
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    if (r == 3) { value[i] = _mon->output; }
    else if (r == 0) { value[i] = _mon->input.i; }
    else if (r == 1) { value[i] = _mon->input.k; }
    else if (r == 2) { value[i] = _mon->input.l; }
    else { status = fmi2Warning; }
  }
  return status;
}

fmi2Status fmi2GetBoolean (fmi2Component _instance,
                           const fmi2ValueReference vr[],
                           size_t nvr, fmi2Boolean value[])
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    if (r < 0) { return fmi2Error; }
    else { status = fmi2Warning; }
  }
  return status;
}

static fmi2String constants[] = {
  "0",
  "1",
  "b", /* 2 */
  "a", /* 3 */
  "c", /* 4 */
  NULL
};

fmi2Status fmi2GetString (fmi2Component _instance,
                          const fmi2ValueReference vr[],
                          size_t nvr, fmi2String value[])
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    if (r == 3) {
      value[i] = RV_value_string(_mon->output);
    }
    else if (r == 1) {
      fmi2Integer v_int = _mon->input.k;
      fmi2String v_str = "";
      if (v_int < 5) { v_str = constants[v_int]; }
      value[i] = v_str;
    }
    else if (r == 2) {
      fmi2Integer v_int = _mon->input.l;
      fmi2String v_str = "";
      if (v_int < 5) { v_str = constants[v_int]; }
      value[i] = v_str;
    }
    else { status = fmi2Warning; }
  }
  return status;
}

fmi2Status fmi2SetReal (fmi2Component _instance,
                        const fmi2ValueReference vr[],
                        size_t nvr, const fmi2Real value[])
{ return fmi2Discard; }

fmi2Status fmi2SetInteger (fmi2Component _instance,
                           const fmi2ValueReference vr[],
                           size_t nvr, const fmi2Integer value[])
{
  struct monitor_fmu *m = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    if (r < 0) { return fmi2Error; }
    else if (r == 0) { m->input.i = value[i]; }
    else if (r == 1) { m->input.k = value[i]; }
    else if (r == 2) { m->input.l = value[i]; }
    else { status = fmi2Discard; }
  }
  return status;
}

fmi2Status fmi2SetBoolean (fmi2Component _instance,
                           const fmi2ValueReference vr[],
                           size_t nvr, const fmi2Boolean value[])
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    if (r == 4) { _mon->soft_reset = value[i]; }
    else { status = fmi2Discard; }
  }
  return status;
}

fmi2Status fmi2SetString (fmi2Component _instance,
                          const fmi2ValueReference vr[],
                          size_t nvr, const fmi2String value[])
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  size_t i = 0;
  fmi2Status status = fmi2OK;
  if (nvr > 0 && NULL == vr)    { return fmi2Error; }
  if (nvr > 0 && NULL == value) { return fmi2Error; }
  for (; i < nvr; i++) {
    fmi2ValueReference r = vr[i];
    const fmi2String   v_str = value[i];
    fmi2Integer        v_int = -1;
    if (strcmp(v_str, "b") == 0) { v_int = 2; }
    else if (strcmp(v_str, "a") == 0) { v_int = 3; }
    else if (strcmp(v_str, "c") == 0) { v_int = 4; }
    if (v_int == -1) {
      v_int = atoi(v_str);
    }
    if (r < 0) { return fmi2Error; }
    else if (r == 1) { _mon->input.k = v_int; }
    else if (r == 2) { _mon->input.l = v_int; }
    else { status = fmi2Discard; }
  }
  return status;
}

fmi2Status fmi2GetFMUstate (fmi2Component _instance, fmi2FMUstate *FMUstate)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  const fmi2CallbackFunctions *functions = _mon->functions;
  int *current_loc;
  if (FMUstate == NULL) { return fmi2Error; }
  if (*FMUstate == NULL) {
    if (functions && functions->allocateMemory) {
      current_loc = (int *)functions->allocateMemory(1, sizeof(int));
      *FMUstate = (fmi2FMUstate)current_loc;
    } else {
      current_loc = (int *)malloc(sizeof(int));
      *FMUstate = (fmi2FMUstate)current_loc;
    }
  } else {
    current_loc = (int *)(*FMUstate);
  }
  *current_loc = _mon->current_loc;
  return fmi2OK;
}

fmi2Status fmi2SetFMUstate (fmi2Component _instance, fmi2FMUstate FMUstate)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  int *current_loc;
  if (FMUstate == NULL) { return fmi2Error; }
  current_loc = (int *)FMUstate;
  _mon->current_loc = *current_loc;
  return fmi2OK;
}

fmi2Status fmi2FreeFMUstate (fmi2Component _instance, fmi2FMUstate *FMUstate)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  const fmi2CallbackFunctions *functions = _mon->functions;
  int *current_loc;
  if (FMUstate == NULL) { return fmi2Error; }
  current_loc = (int *)(*FMUstate);
  *FMUstate = NULL;
  if (current_loc == NULL) { return fmi2OK; }
  if (functions && functions->freeMemory) {
    functions->freeMemory(current_loc);
  } else {
    free(current_loc);
  }
  return fmi2OK;
}

fmi2Status fmi2SerializedFMUstateSize (fmi2Component _instance,
                                       fmi2FMUstate FMUstate, size_t *size)
{ return fmi2Warning; }

fmi2Status fmi2SerializeFMUstate (fmi2Component _instance,
                                  fmi2FMUstate FMUstate,
                                  fmi2Byte serializedState[], size_t size)
{ return fmi2Warning; }

fmi2Status fmi2DeSerializeFMUstate (fmi2Component _instance,
                                    const fmi2Byte serializedState[],
                                    size_t size, fmi2FMUstate* FMUstate)
{ return fmi2Warning; }

fmi2Status fmi2GetDirectionalDerivative (fmi2Component _instance,
                                         const fmi2ValueReference vUnknown_ref[],
                                         size_t nUnknown,
                                         const fmi2ValueReference vKnown_ref[],
                                         size_t nKnown,
                                         const fmi2Real dvKnown[],
                                         fmi2Real dvUnknown[])
{ return fmi2Warning; }

/* Functions for FMI for Co-Simulation */
fmi2Status fmi2SetRealInputDerivatives (fmi2Component _instance,
                                        const fmi2ValueReference vr[], size_t nvr,
                                        const fmi2Integer order[],
                                        const fmi2Real value[])
{ return fmi2Discard; }

fmi2Status fmi2GetRealOutputDerivatives (fmi2Component _instance,
                                         const fmi2ValueReference vr[], size_t nvr,
                                         const fmi2Integer order[], fmi2Real value[])
{ return fmi2Warning; }

fmi2Status fmi2CancelStep (fmi2Component _instance) { return fmi2Warning; }

fmi2Status fmi2DoStep (fmi2Component _instance, fmi2Real currentCommunicationPoint,
                       fmi2Real communicationStepSize,
                       fmi2Boolean noSetFMUStatePriorToCurrentPoint)
{
  struct monitor_fmu *_mon = (struct monitor_fmu *)_instance;
  monitor_input_t *input = &(_mon->input);
  int *current_loc = &(_mon->current_loc);
  monitor_mask_t *mask = NULL;
  RV_reset reset = NO_RESET;
  RV_value output;

  if (_mon->soft_reset) {
    reset = SOFT_RESET;
    _mon->soft_reset = 0;
  }
  output = monitor_scalar(input, mask, reset, current_loc);
  if (output <= RV_ERROR) {
    _mon->output = output;
    return fmi2OK;
  } else {
    return fmi2Error;
  }
}

fmi2Status fmi2GetStatus (fmi2Component _instance, const fmi2StatusKind s, fmi2Status *value)
{
  return fmi2Warning;
}

fmi2Status fmi2GetRealStatus(fmi2Component _instance, const fmi2StatusKind s, fmi2Real *value)
{
  return fmi2Warning;
}

fmi2Status fmi2GetIntegerStatus (fmi2Component _instance,
                                 const fmi2StatusKind s,
                                 fmi2Integer *value)
{
  return fmi2Warning;
}

fmi2Status fmi2GetBooleanStatus (fmi2Component _instance,
                                  const fmi2StatusKind s,
                                 fmi2Boolean *value)
{
  return fmi2Warning;
}

fmi2Status fmi2GetStringStatus (fmi2Component _instance,
                                const fmi2StatusKind s,
                                fmi2String *value)
{
  return fmi2Warning;
}

